require 'yaml'
require 'log4r'
include Log4r

require 'iowa/Constants'
require 'iowa/Extensions/AllExtensions'
require 'iowa/Association'
require 'iowa/Context'
require 'iowa/Element'
require 'iowa/Tag'
require 'iowa/DynamicElements'
require 'iowa/Form'
require 'iowa/Component'
require 'iowa/ComponentProxy'
require 'iowa/TemplateParser'
require 'iowa/BindingsParser'
require 'iowa/KeyValueCoding'
require 'iowa/PageStore'
require 'iowa/Session'
require 'iowa/Application'
require 'iowa/Request'
require 'iowa/Email'
require 'iowa/PrettyException'

require 'socket'
require 'resolv'

module Iowa

	@@app = nil

	# An Iowa appplication's configuration is given in a plain text file
	# formatted as YAML (http://www.yaml.org).

	def Iowa.readConfiguration(configuration_file)
		begin
			cf = File.open configuration_file

			rescue Exception => exception
				puts "Failure while accessing the configuration file (#{configuration_file})\n"
				raise
		end

		@@config = YAML::load(cf).downcase_keys

		checkConfiguration
	end


	# Forks the process.  The parent outputs the PID of the child (so that
	# an external invocation process can capture this information) and then
	# exits.  The child sets inself as it's session leader and returns.

	def Iowa.startDaemon(configuration_file)
		readConfiguration(configuration_file)
		startLogging

		begin
			#if (fork)
			#	exit!
			#end

			#Process.setsid

			if (child_pid = fork)
				puts "PID #{child_pid}"
				exit!
			end

			Process.setsid

			mylog = Logger['iowa_log']
			mylog.info "Process forked and daemonized as PID #{Process.pid}."
		rescue Exception
			mylog = Logger['iowa_log']
			mylog.info "Platform (#{RUBY_PLATFORM}) does not appear to support fork/setsid; skipping"
		end
	end

	def Iowa.configureExceptionScreens
		@exception_screens = []
		@exception_hosts = []
		if @@config.has_key?('exceptions')
			if @@config['exceptions'].has_key?('screens')
				@@config['exceptions']['screens'].each do |item|
					new_item = {}
					next unless item.is_a?(Hash)
					item.each do |k,v|
						k.sub!(/^\s+/,'')
						v.sub!(/^\s+/,'')
						if m = /^\/(.*)\//.match(k)
							new_item[Regexp.new(m[1]),true] = v
						else
							new_item[k] = v
						end
					end
					@exception_screens.push(new_item)
				end
			end
			if @@config['exceptions'].has_key?('hosts')
				@@config['exceptions']['hosts'].each do |item|
					new_item = {}
					next unless item.is_a?(Hash)
					item.each do |k,v|
						k.sub!(/^\s+/,'')
						if v.is_a?(Array)
							list = []
							v.each {|x| list.push(Regexp.new(x.sub(/^\s+/,''),true))}
							new_item[k] = list
						end
					end
					@exception_hosts.push(new_item)
				end
			end
		end
	end

	# Check that the minimum required information is present in the
	# configuration, and set the defaults for information that was
	# omitted. Exceptions will be raised if socket information is
	# missing, or if either socket or logging information seems to
	# have errors.

	def Iowa.checkConfiguration
		# Make sure socket info was provided
		raise "A socket specification must be provided in the configuration file." unless @@config.has_key? 'socket'
		raise "Iowa supports both TCP and Unix domain sockets.  Please specify the socket to use within the 'socket' section of the configuration file." unless (@@config['socket'].class.to_s == 'Hash' and (@@config['socket'].has_key? 'path' or @@config['socket'].has_key? 'port'))
		raise "Please define only one of a TCP or a Unix domain socket in the configuration file." if @@config['socket'].has_key? 'path' and @@config['socket'].has_key? 'port'

		# Verify that the socket information is valid.  i.e. that the provided
		# port is a number, that the hostname for a TCP socket doesn't have
		# funny characters in it, or that the path for a Unix socket is writeable.
		#raise "A port number must be specified as an integer." if @@config['socket'].has_key? 'port' and @@config['socket']['port'].class.to_s == 'Fixnum'
		raise "The socket hostname contains illegal characters." unless @@config['socket'].has_key? 'hostname' and @@config['socket']['hostname'] =~ /^[a-zA-Z0-9\.\-]*$/
		if @@config['socket'].has_key? 'path'
			begin
				File.unlink @@config['socket']['path'] if FileTest.exist? @@config['socket']['path']
				sf = File.open(@@config['socket']['path'],'w')
				@@config['socket']['basename'] = File.basename(@@config['socket']['path'])
				sf.close
				File.unlink @@config['socket']['path']
			rescue Exception => exception
				puts "The path for a Unix socket must be writeable."
				raise
			end
		end

		# If any logging information has been specified, that needs to be
		# checked, as well.

		if @@config.has_key? 'logging'
			if @@config['logging'].has_key? 'basedir'
				raise "The logging basedir (#{@@config['logging']['basedir']}) must be a directory." unless FileTest.directory? @@config['logging']['basedir']
				raise "The logging basedir #{@@config['logging']['basedir']}) must be writeable." unless FileTest.writable? @@config['logging']['basedir']
			end
		
			raise "The minimum logging level (minlevel) must be an integer between 0 to 7, inclusively." if @@config['logging'].has_key? 'minlevel' and (@@config['logging']['minlevel'] < 0 or @@config['logging']['minlevel'] > 7)

			raise "The maximum log file size (maxsize) must be greater than zero." if @@config.has_key? 'maxsize' and @@config['maxsize'] <= 0

			raise "The maximum log file age (maxage) must be greater than zero." if @@config.has_key? 'maxage' and @@config['maxage'] <= 0
		end

		# Now apply any defaults for optional items which were not specified.

		@@config['logging'] = Hash.new unless @@config.has_key? 'logging'
		@@config['logging']['basedir'] = '/tmp' unless @@config['logging'].has_key? 'basedir'
		@@config['logging']['minlevel'] = 'info' unless @@config['logging'].has_key? 'minlevel'
		@@config['logging']['maxsize'] = 1024 * 1024 unless @@config['logging'].has_key? 'maxsize'
		@@config['logging']['maxage'] = 86400 unless @@config['logging'].has_key? 'maxage'
		if @@config['socket'].has_key? 'path'
			@@config['logging']['filename'] = "#{@@config['logging']['basedir']}/#{@@config['socket']['basename']}.log"
		else
			@@config['logging']['filename'] = "#{@@config['logging']['basedir']}/#{@@config['socket']['hostname']}:#{@@config['socket']['port']}.log"
		end

		configureExceptionScreens
	end

	# Initialize logging mechanisms

	def Iowa.startLogging
		lcfg = @@config['logging']
		mylog = Logger.new 'iowa_log'
		myformat = PatternFormatter.new :pattern => '[%l] %C @ %d :: %M'
		mybaselog = RollingFileOutputter.new 'baselog', :maxsize => lcfg['maxsize'], :maxtime => lcfg['maxage'], :level => lcfg['minlevel'], :filename => lcfg['filename'], :trunc => false, :formatter => myformat
		mylog.outputters = mybaselog

#		myApacheLikeLog = Logger.new 'iowa_apachelike_log'
#		myApacheLikeFormat = PatternFormatter.new :pattern => '%M'
#		mybaseApacheLikeLog = RollingFileOutputter.new 'apachelike_baselog', :maxsize => 999999999, :maxtime => 86400, :level => lcfg['minlevel'], :filename => lcfg['filename'] + '.apachelike', :trunc => false, :formatter => myApacheLikeFormat
#		myApacheLikeLog.outputters = mybaseApacheLikeLog

		mylog.info 'Logging subsystem initialized.'
	end

	def Iowa.maybe_kind_of(obj,classname)
		r = false
		obj.class.ancestors.each {|a| r ||= a.to_s == classname}
		r
	end
	
	# Check the exception against the list of exception screens, if any.  Return as
	# soon as a match is found.

	def Iowa.checkScreen(exception)
		klass = exception.class.to_s
		@exception_screens.each do |item|
			item.each do |k,v|
				if k.is_a?(Regexp)
					if k.match(klass)
						return v
					end
				elsif maybe_kind_of(exception,k)
					return v
				end
			end
		end
		nil
	end

	def Iowa.internal_host?(request)
		addr = false
		browser = false
		@exception_hosts.each do |host|
			if host.has_key?(Caddresses)
				host[Caddresses].each do |a|
					if a.match(request.remote_host)
						addr = true
						break
					end
				end
			elsif host.has_key?(Cips)
				host[Cips].each do |i|
					if i.match(request.remote_addr)
						addr = true
						break
					end
				end
			else
				addr = true
			end
			if host.has_key?(Cbrowsers)
				host[Cbrowsers].each do |b|
					if b.match(request.headers_in[CUserAgent])
						browser = true
						break
					end
				end
			else
				browser = true
			end
		end
		(addr and browser)
	end

	def Iowa.handleRequest(request)
		deployed_screen = false
		specific_page = nil
		buffer = nil
		original_exception = nil
		loop do
			request.content_type = Ctext_html
			buffer = ''
			context = nil
			exception = nil
			exception = catch(:session_error) do
				begin
					context = Iowa::Context.new(request, buffer)
					context.prior_exception = original_exception if original_exception
					app.handleRequest(context,specific_page)
				rescue Exception => exception
				end
			end
			if exception.to_s != C_empty and exception.kind_of?(Exception)
				mylog = Logger['iowa_log']
				mylog.warn "Execution Error: #{exception} :: " + exception.backtrace.join(".....").to_s unless original_exception
				if internal_host?(request)
					mylog.info "Showing PrettyException"
					buffer << Iowa::PrettyException.new(exception).to_s
				else
					if deployed_screen == false and specific_page = checkScreen(exception)
						mylog.info "Deploying exception screen #{specific_page}"
						request.status_line = '500 Internal Server Error'
						original_exception = exception
						deployed_screen = true
						retry
					else
						mylog.info "Showing ugly exception"
						request.status_line = '500 Internal Server Error'
						exception = original_exception if original_exception
						buffer << "<p> An unhandled error has occured.  Please inform your system or your site's administrator of the error.</p>"
						buffer << "<p>#{exception}<br/>"
						buffer << exception.backtrace.join("<br/>\n").to_s
						buffer << "</p>"
						status = 'EOUT'
					end
				end
			end
			break
		end
		buffer
	end

	# Handle the communications coming in on the monitored socket,
	# create a context object from the data received, and then pass
	# the context information into the Application object for final
	# handling.  Exception handling is simply via capturing the Exception
	# and outputting a stack backtrace (this could be improved).

	def Iowa.handleConnection(socket)
		begin
			Thread.current.priority = 2
			start_time = Time.now
			mylog = Logger['iowa_log']
			# Having two logfiles with much the same data in them is stupid.  Fix this.
			message = ''
			while (recv = socket.recv(8192)) != C_empty
				message << recv
			end
			socket.shutdown(0)
			read_time = Time.now
			status = COK
		rescue Exception => exception
			mylog.error "Failure While Reading Inbound Data: #{exception}"
			status = CEIN
		end

		begin
			request = Marshal.load(message)
		rescue Exception => exception
			mylog.error "Inbound Data Corruption: #{exception}"
			status = CEIN
		end

		buffer = handleRequest(request)

		begin
			request.content = buffer
			message = Marshal.dump(request)
			socket.write(message)
			socket.flush
			socket.shutdown(1)
			end_time = Time.now
		rescue Exception => exception
			mylog.error "Failure While Writing Outbound Data: #{exception}"
		end

#		vhost = nil
#		request.unparsed_uri =~ /http:\/\/([^\/]*)\//
#		vhost = $1
#		referer = request.headers_in['Referer'] != '' ? request.headers_in['Referer'] : request.uri
#		myapachelikelog.info "#{request.remote_host} #{request.hostname} - [#{Time.now.strftime('%d/%b/%Y:%H:%M:%S')}] #{buffer.length} \"#{request.request_method} #{request.unparsed_uri} HTTP/1.1\" 200 #{buffer.length} \"#{referer}\" \"#{request.headers_in['User-Agent']}\" \"#{request.headers_in['Cookie']}\""
		logline = "#{start_time} (#{read_time.to_f - start_time.to_f}/#{end_time.to_f - start_time.to_f}) :: #{request.uri} #{status} #{buffer.length}B"
		mylog.info logline
	end

	# Outputs the location of the socket being monitored, then enters the
	# event loop to wait for and handle connections.

	def Iowa.eventLoop(server)
		Logger['iowa_log'].info 'Entering the main event loop...'
		loop do
			socket = server.accept
			Thread.start {handleConnection socket}	
		end
	end

	# Outputs the location of the socket being monitored, then enters the
	# event loop to wait for and handle connections.

	def Iowa.to_hex(n)
		sprintf('%02s',n.to_i.to_s(16)).sub(' ','0')
	end
	
	def Iowa.run()
		mylog = Logger['iowa_log']
		my_ip = @@config['socket']['hostname']
		my_ip_hex = my_ip.split('.',4).collect {|x| to_hex(x)}.join

		if @@config['socket'].has_key? 'port'
			socket_host = @@config['socket']['hostname']
			socket_port = @@config['socket']['port']
			my_port_hex = sprintf('%04s',socket_port.to_i.to_s(16)).gsub(' ','0')
			begin
				# This should be configurable and runtime toggleable.
				TCPServer.do_not_reverse_lookup = true
				if socket_host.to_s == ''
					server = TCPserver.new(socket_port)
					mylog.info "Listening at localhost:#{socket_port}."
				else
					server = TCPserver.new(socket_host,socket_port)
					mylog.info "Listening at #{socket_host}:#{socket_port}."
				end
				Iowa.app.location = "#{my_ip_hex}#{my_port_hex}"
			rescue Exception => exception
				mylog.fatal "Failure while attempting to establish a TCP socket at #{socket_host}:#{socket_port} : #{exception}"
				raise
			end
		else
			begin
				server = UNIXServer.new(@@config['socket']['path'])
				File.chmod(0666, @@config['socket']['path'])
				mylog.info "Listening on #{@@config['socket']['path']}."
			rescue Exception => exception
				mylog.fatal "Failure while attemting to establish a Unix socket at #{@@config['socket']['path']} : #{exception}"
				raise
			end
			if Iowa.app.serial_number.to_s != ''
				Iowa.app.location = "00000000#{sprintf('%04s',Iowa.app.serial_number.to_i.to_s(16)).gsub(' ','0')}"
			else
				Iowa.app.location = ''
			end
		end

		# initialLoad performs the first loading of the application files.
		# This has been seperated from reloadModified in order to
		# potentially provide a hook that can be used to improve
		# overall loading performance.  If we just could persistently
		# cache a parsed template....
		app.initialLoad()

		begin
			eventLoop(server)
		rescue Exception => exception
			mylog.fatal "Catastrophic failure in main event loop: #{exception} :: " + exception.backtrace.join(".....").to_s
		ensure
			server.close
			File.delete(@@config['socket']['path']) if @@config['socket'].has_key? 'path'
		end
	end

	def Iowa.app
		if @@app
			@@app
		else
			@@app = Application.newApplication(Dir.getwd)
		end
	end

end
